{
    This file is part of the Cheb's Game Engine,
    Copyright (c) 2004-2006 by Anton Rzheshevski (chebmaster@mail.ru),
      and contains source of the demo #001.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 **********************************************************************}

unit _demo001_scene;

interface

 uses
  SysUtils,
  {$ifdef fpc}
    un_typedefs in '../../cge/src/un_typedefs.pas',
    mo_hub in '../../cge/src/mo_hub.pas',
    mo_classes in '../../cge/src/mo_classes.pas',
    mo_textures in '../../cge/src/mo_textures.pas',
    mo_module in '../../cge/src/mo_module.pas',
    mo_controls in '../../cge/src/mo_controls.pas'
  {$else}
    un_typedefs,
    mo_hub,
    mo_classes,
    mo_textures,
    mo_module,
    mo_controls
  {$endif}
  , md2, math;
  
Const
  FOV = 90;
  NEARPLANE = 1;
  FARPLANE = 500;
  ASPECT = 4/3;
  
Type
  T8array = array[0..7] of glUint;

  { TDemo001Scene }

  TDemo001Scene = class
  public
    tex_pinky,
    tex_skull,
    tex_walls: glUint;
    tex_fire: array[0..8] of gluint;
    skull,
    flame,
    pinky,
    box: TMd2;
    constructor Create;
    procedure Render(Brightness: float);
    destructor Destroy; override;
  end;
  
implementation


{ TDemo001Scene }

procedure _LoadTex(tex: glUint; FileName: string);
var
  mode: glUint;
  t:  integer;
  m: TImageMode;
  w,h, stub: integer;
  p: pointer;
  fullname: AnsiString;
begin
  fullname:=Module.BaseDir + 'content/' + FileName;
  if VerboseLog then addlog('Loading "%0" ... ', [FullName]);
  Filename:='$' + Filename;
  CreateContainerFromFile(PChar(FileName), PChar(fullname), stub);
  PrepareToDecodePic (PChar(FileName), w, h, m);
  GetMem(p, w * h * ImgModeBpp[m] + 16);
  Try
    DecodePic(p);
    glBindTexture(GL_TEXTURE_2D, tex);
    if m = Im_RGBA then mode:=GL_RGBA else mode:=GL_RGB;
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D (GL_TEXTURE_2D, 0, mode, w, h, 0, mode, GL_UNSIGNED_BYTE, p);
  Finally
    FreeMem(p);
    DeleteContainer(PChar(FileName));
  End;
  if VerboseLog then AddLogOk;
end;

constructor TDemo001Scene.Create;
var i: integer;
begin
  glGenTextures(1, @tex_pinky);
  glGenTextures(1, @tex_skull);
  glGenTextures(1, @tex_walls);
  for i:=0 to 7 do
    glGenTextures(1, @tex_fire[i]);
  for i:=0 to 7 do
    _LoadTex(tex_fire[i], 'Flame_Red_' + IntToStr(i) + '.png');
  _LoadTex(tex_pinky, 'Skindemon.png');
  _LoadTex(tex_skull, 'lostsoul_Skin.png');
  _LoadTex(tex_walls, '005roche.png');
  skull:=TMD2.Create(BaseDir + Module.BaseDir + 'content/lostsoul.MD2');
  flame:=TMD2.Create(BaseDir + Module.BaseDir + 'content/lostsoul_flame.MD2');
  pinky:=TMD2.Create(BaseDir + Module.BaseDir + 'content/pinky.MD2');
  box:=TMD2.Create(BaseDir + Module.BaseDir + 'content/box.MD2');
end;

destructor TDemo001Scene.Destroy;
var i: integer;
begin
  glDeleteTextures(1, @tex_pinky);
  glDeleteTextures(1, @tex_skull);
  glDeleteTextures(1, @tex_walls);
  for i:=0 to 7 do glDeleteTextures(1, @tex_fire[i]);
  skull.free;
  pinky.free;
  flame.free;
  box.free;
end;


procedure TDemo001Scene.Render(Brightness: float);
var
  flameframe: integer;
  angle: float;
  top, bottom, left, right: glFloat;
  SkullBrightness, FlameBrightness: float;
begin
  Try
    glViewport(0, 0, DisplayWidth, DisplayHeight);
    glMatrixMode(GL_PROJECTION);        // Change Matrix Mode to Projection
    glLoadIdentity();                   // Reset View
    top := tan (FOV*3.14159/360.0)*NEARPLANE;
    bottom := -top;
    left := ASPECT*bottom;
    right := ASPECT*top;
    glFrustum (left, right, bottom, top, NEARPLANE, FARPLANE);
    gltranslatef(0,-50,-100);
    
    //glClear(GL_DEPTH_BUFFER_BIT);

    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    glCullFace(GL_FRONT);
    glEnable(GL_CULL_FACE);
{glActiveTextureARB (GL_TEXTURE1_ARB);
glDisable (GL_TEXTURE_2D);

glActiveTextureARB (GL_TEXTURE0_ARB);
glEnable (GL_TEXTURE_2D);
glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_PRIMARY_COLOR);
glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_REPLACE);}




    flameframe:=(Tick() div 100) and $7;
    angle:=tick()/100;
    if Brightness < 1 then FlameBrightness:=1
                      else FlameBrightness:=sqrt(1/Brightness);
    if Brightness < 1 then SkullBrightness:=(Brightness + FlameBrightness)/2
                      else SkullBrightness:=1;

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glColor3f(Brightness, Brightness, Brightness);
    glBindTexture(GL_TEXTURE_2D, tex_walls);
    glrotatef(angle, 0, 1, 0);
    glPushMatrix;
    glScalef(2,2,2);
    glDepthFunc(GL_ALWAYS);
    box._Draw(0);
    glDepthFunc(GL_LEQUAL);
    glBindTexture(GL_TEXTURE_2D, tex_pinky);
    glPopMatrix;
    glPushmatrix;
    gltranslatef(0,0,-30);
    glScalef(0.5,0.5,0.5);
    pinky._Draw(0);

    glPopMatrix;
    gltranslatef(0, 50, 30);
    glrotatef(- 180, 0, 1, 0);
    glScalef(0.8,0.8,0.8);
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GREATER, 0.5);
    glColor3f(SkullBrightness, SkullBrightness, SkullBrightness);
    glBindTexture(GL_TEXTURE_2D, tex_skull);
    skull._Draw(0);
    glEnable(GL_BLEND);
    glAlphaFunc(GL_GREATER, 0);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);//GL_DST_COLOR, GL_SRC_ALPHA);//

    glColor3f(FlameBrightness,FlameBrightness,FlameBrightness);
    glBindTexture(GL_TEXTURE_2D, tex_fire[flameframe]);
    flame._Draw(0);
  Except
    Die('Scene render crashed');
  End;
end;

end.
